package ic;
import jv.vecmath.*;
import jv.geom.*;
import jdsl.core.ref.*;
import jdsl.core.api.*;
import java.util.*;

public class Face extends NodeTree{

  public static final int TRIANGLE = 3;
  public static final int SQUARE = 4;
  public static final int PENTAGON = 5;
  public static final int HEXAGON = 6;
  public static final int HEPTAGON = 7;
  public static final int OCTAGON = 8;
  public static final int ENEAGON = 9;
  public static final int DECAGON = 10;
  public static final int DODECAGON = 12;
  public static final int ICOSAGON = 20;


  protected int numVertices = 0; //Number of vertices this face
  protected PdVector A, B;
  protected double theta; // thetha is the diedral angle.
  protected PgPointSet myVertices = new PgPointSet(3); //Holds the coordinates of vertices this face when it is plan
  protected PgPointSet myCoordsVertices = new PgPointSet(3);
  protected PdMatrix rotateMatrix;
  protected int myIndex = 0; //Index this face
  private double percentRotation = 0.0; //Percentual de rotao
  private HashSet polygons = new HashSet();


//Constructors

  public Face(){
    super();
    //this(3,new PdVector(0.0,0.0,0.0), new PdVector(5.0,0.0,0.0), Math.PI/2, 0);
  }

  public Face(int numVertices){
    //this(numVertices,new PdVector(0.0,0.0,0.0), new PdVector(5.0,0.0,0.0), Math.PI/2);
    this.numVertices=numVertices;
  }

  public Face(int numVertices, PdVector A,PdVector B, double theta, int my_Index){
    super();
    this.numVertices = numVertices;
    this.A = A;
    this.B = B;
    this.theta = theta;
    this.myIndex = my_Index;
    generateCoordsMyVertices(numVertices, A, B);
  }

//End of Constructors


  private void generateCoordsMyVertices(int numVertices, PdVector P, PdVector Q){
    PdVector M,N,R;
    R = new PdVector(3);
    double RotateAngle = Math.PI*(numVertices-2)/numVertices;
    M = PdVector.copyNew(P);
    N = PdVector.copyNew(Q);
    this.myVertices.addVertex(PdVector.copyNew(M)); this.myCoordsVertices.addVertex(PdVector.copyNew(M));
    this.myVertices.addVertex(PdVector.copyNew(N)); this.myCoordsVertices.addVertex(PdVector.copyNew(N));

    for (int i=0;i<numVertices-2;i++){
        PuVectorGeom.rotatePointAroundLine(R,M,N,new PdVector(0,0,1),RotateAngle);
        this.myVertices.addVertex(PdVector.copyNew(R));this.myCoordsVertices.addVertex(PdVector.copyNew(R));
        M = PdVector.copyNew(N);
        N = PdVector.copyNew(R);
    }
}

  public Position returnPositionFace(int index, Face face){
    PreOrderIterator iter = new PreOrderIterator(face);
    try {
      return this.returnPositionOfIterator(index, iter);
    }
    catch (Exception exp){
      System.out.println("Face "+ Integer.toString(index) +" not found");
      return face.root();
    }

  }

  private Position returnPositionOfIterator(int index , PreOrderIterator iter) throws Exception{
    Face currentFace;
    Position currentPosition;

    while(iter.hasNext()){
    currentPosition = iter.nextPosition();
    currentFace = ((Face) currentPosition.element());
    if(currentFace.myIndex==index)
      return currentPosition;
    }
    throw new Exception("Face "+ Integer.toString(index) +" not found");

  }

  public Position returnPositionFace(int index, Tree solidTree){
    PreOrderIterator iter = new PreOrderIterator(solidTree);
    try {
      return this.returnPositionOfIterator(index, iter);
    }
    catch (Exception exp){
      System.out.println("Face "+ Integer.toString(index) +" not found");
      return solidTree.root();
    }
  }

  public PdVector getVertex(int index){
    return myVertices.getVertex(index);
  }

  public void setPercentRotation(double perc){
    this.percentRotation = perc;
  }

  public double getPercentRotation(){
    return this.percentRotation;
  }

  private double partialAngle(){
    return (theta*percentRotation);
  }
  public void updateCoordsVertices(double perc, PdMatrix parentMatrix){
    this.percentRotation = perc;
//    this.myCoordsVertices;
  }

  public void updateCoordsVertices(PdMatrix parentFaceMatrix){
    PdVector v;
    updateRotateMatrix(parentFaceMatrix);

    //if (this.myCoordsVertices.getNumVertices()>0)
      this.myCoordsVertices.removeAllVectorFields();
    this.myCoordsVertices.setNumVertices(this.myVertices.getNumVertices());

    for(int i = 0; i < this.myVertices.getNumVertices();i++){
      v = PdVector.copyNew(this.myVertices.getVertex(i));
      //v = this.myVertices.getVertex(i);
      v.addEntry(1.0); // Adiciona mais uma coordenada no fim do vetor para que ele tenha dimenso 4
      v.leftMultMatrix(this.rotateMatrix);
      v.setEntry(0,v.getEntry(0)/v.getEntry(3));
      v.setEntry(1,v.getEntry(1)/v.getEntry(3));
      v.setEntry(2,v.getEntry(2)/v.getEntry(3));
      //this.myCoordsVertices.addVertex(v);
      this.myCoordsVertices.setVertex(i,PdVector.copyNew(v));

    }



  }
  private void updateRotateMatrix(PdMatrix intermediateMatrix){
    PdVector rotateAxis = PdVector.subNew(this.B,this.A); //Calcula o vetor de rotao v=B-A
    //rotateAxis.normalize();//Normaliza o vetor de rotao
    //rotateAxis.multScalar(10);//Multiplicao vetor de rotao normalizado por 10, pois o mtodo rotateLine() no aceita vetores normalizados
    this.rotateMatrix = PuReflect.rotateLine(this.A,rotateAxis,this.partialAngle()); //Obtem a matriz de rotao da face atual
    //o ngulo de rotao parcial  retornado pelo mtodo partialAngle()
    this.rotateMatrix.leftMult(intermediateMatrix);//Executa a multiplicao de matrizes rotateMatrix = rotateMatrix * intermediateMatrix
  }

  public PdMatrix getRotateMatrix(){
    return this.rotateMatrix;
  }

  public static void rotateSubFaces(double perRotation, Tree solidTree){
  Face currentFace, parentFace;

  PdMatrix mainMatrix = new PdMatrix(4);
  mainMatrix.setIdentity();

  Face mainFace = (Face)solidTree.root().element();
  PreOrderIterator iter = new PreOrderIterator(solidTree);

  mainFace.setPercentRotation(perRotation);
  mainFace.rotateMatrix = mainMatrix; //Configura a matriz de rotao da face me como sendo a matriz Identidade
  //mainFace.myCoordsVertices = mainFace.myVertices;

  iter.nextPosition(); //Pula a face me (raiz)

  while(iter.hasNext()){
    currentFace = ((Face)iter.nextPosition().element());
    parentFace = ((Face)solidTree.parent(iter.position()).element());
    currentFace.setPercentRotation(perRotation);
    currentFace.updateCoordsVertices(parentFace.getRotateMatrix());
  }

}

  public void rotateSubFaces(double perRotation) {
    Face currentFace, parentFace;

    PdMatrix mainMatrix = new PdMatrix(4);
    mainMatrix.setIdentity();

    Face mainFace = (Face) this.root().element();
    PreOrderIterator iter = new PreOrderIterator(this);

    mainFace.setPercentRotation(perRotation);
    mainFace.rotateMatrix = mainMatrix; //Configura a matriz de rotao da face me como sendo a matriz Identidade//mainFace.myCoordsVertices = mainFace.myVertices;

    iter.nextPosition(); //Pula a face me (raiz)

    while (iter.hasNext()) {
      currentFace = ( (Face) iter.nextPosition().element());
      parentFace = ( (Face) this.parent(iter.position()).element());
      currentFace.setPercentRotation(perRotation);
      currentFace.updateCoordsVertices(parentFace.getRotateMatrix());
    }
  }


  public int getNumVertices(){ return numVertices; }
  public int getMyIndex() { return this.myIndex; }
  public String toString() { return Integer.toString(numVertices);}
  public HashSet getTypeFaces(){
    Face currentFace;
    PreOrderIterator iter = new PreOrderIterator(this);
    while (iter.hasNext()) {
      currentFace = ( (Face) iter.nextPosition().element());
      this.polygons.add(new Integer(currentFace.getNumVertices()));
    }
    return this.polygons;
  }

}//End of Class Face